import redis
from .settings import *


class Redis:

    redis_conf = REDIS['default']
    redis_host = redis_conf['HOST']
    redis_port = redis_conf['PORT']
    redis_pool = redis.ConnectionPool(
        host=redis_host, port=redis_port, db=0)

    @classmethod
    def get_client(cls):
        return redis.Redis(connection_pool=cls.redis_pool)


class Client:
    """
    Used as interface in all other modules.
    """
    instance = Redis  # overwrite this for mocking

    @classmethod
    def get_client(cls):
        return cls.instance.get_client()



import unittest
from ..services import Redis

from rorm.client import Client
from rorm.types import Hash


class HashTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        Client.instance = cls.service = Redis('test-service-hash')
        cls.client = cls.service.get_client()
        cls.client.flushall()

    @classmethod
    def tearDownClass(cls):
        cls.client.flushall()
        cls.service.stop()

    def setUp(self):
        self.dragon_red = Hash()  # uses auto generated key
        self.dragon_black = Hash(
            key='custom-key-3442-24 423@#&^$!_*?><|\;')

    def test_hash_with_auto_key(self):
        import uuid
        self.assertIsInstance(self.dragon_black.key, str)
        self.assertEqual(len(self.dragon_black.key), len(str(uuid.uuid4())))
        self.assertNotEqual(self.dragon_black.key, self.dragon_red.key)

    def test_hash_with_custom_key(self):
        self.assertEqual(
            self.dragon_black.key, 'custom-key-3442-24 423@#&^$!_*?><|\;')

    def test_hash_wrong_initial_data(self):
        self.assertRaises(ValueError, Hash, tuple)
        self.assertRaises(ValueError, Hash, 4)

    def test_type_of_object(self):
        self.assertIsInstance(self.dragon_red, Hash)

    def test_length_of_object(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344})

        self.assertEqual(len(dragon_blue), 3)

    def test_get_string_data(self):
        self.dragon_red['name'] = 'Dormith Champion Of The Red'
        self.dragon_black['name'] = 'Jeruth Lord Of The Black'
        self.dragon_red['element'] = 'fire'

        self.assertEqual(
            self.dragon_red['name'], b'Dormith Champion Of The Red')
        self.assertEqual(
            self.dragon_black['name'], b'Jeruth Lord Of The Black')
        self.assertEqual(self.dragon_red['element'], b'fire')

    def test_get_int_data(self):
        self.dragon_red['age'] = 23489
        self.dragon_black['age'] = 89239
        self.assertEqual(self.dragon_red['age'], b'23489')
        self.assertEqual(self.dragon_black['age'], b'89239')

    def test_get_raw_data(self):
        self.dragon_red['real_name'] = '!@#$%^&*()_+{}:"<>?`\`"'
        self.dragon_black['real_name'] = '!()00&*()_+{}:"<>?`\`"'
        self.assertEqual(
            self.dragon_red['real_name'], b'!@#$%^&*()_+{}:"<>?`\`"')
        self.assertEqual(
            self.dragon_black['real_name'], b'!()00&*()_+{}:"<>?`\`"')

    def test_clear_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344})

        self.assertEqual(len(dragon_blue), 3)
        dragon_blue.clear()
        self.assertEqual(len(dragon_blue), 0)
        self.assertEqual(dragon_blue, {})

    def test_items_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )

        self.assertCountEqual(
            dragon_blue.items(),
            ((b'name', b'Lusarth The Powerful One'),
                (b'element', b'ice'),
                (b'age', b'123344'))
        )

    def test_keys_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )

        self.assertCountEqual(
            dragon_blue.keys(),
            (b'name', b'element', b'age',))

    def test_pop_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )

        self.assertEqual(
            dragon_blue.pop('name'),
            b'Lusarth The Powerful One'
        )
        self.assertEqual(
            dragon_blue.pop('element'),
            b'ice'
        )
        self.assertEqual(
            dragon_blue.pop('age'),
            b'123344'
        )

    def test_popitem_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )
        self.assertCountEqual(
            (
                dragon_blue.popitem(),
                dragon_blue.popitem(),
                dragon_blue.popitem()
            ),
            (
                (b'name', b'Lusarth The Powerful One'),
                (b'age', b'123344'),
                (b'element', b'ice')
            )
        )

        self.assertRaises(
            KeyError,
            dragon_blue.popitem
        )

    def test_setdefault_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )
        self.assertEqual(
            dragon_blue.setdefault('weight', 'heavy'),
            'heavy'
        )
        self.assertDictEqual(
            dict(dragon_blue),
            {
                b'name': b'Lusarth The Powerful One',
                b'weight': b'heavy',
                b'element': b'ice',
                b'age': b'123344'
            }
        )

    def test_update_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )

        dragon_blue.update({'height': 'big', 'length': 'long'})

        self.assertDictEqual(
            dict(dragon_blue),
            {
                b'name': b'Lusarth The Powerful One',
                b'element': b'ice',
                b'age': b'123344',
                b'height': b'big',
                b'length': b'long'
            }
        )

    def test_values_method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )
        self.assertCountEqual(
            dragon_blue.values(),
            (b'Lusarth The Powerful One', b'ice', b'123344')
        )

    def test___str___method(self):
        dragon_blue = Hash({
            'name': 'Lusarth The Powerful One',
            'element': 'ice',
            'age': 123344}
        )

        self.assertCountEqual(
            dragon_blue.__str__(),
            "{b'name': b'Lusarth The Powerful One', "
            "b'element': b'ice', "
            "b'age': b'123344'}"
        )



import unittest
import string
import itertools

from factory.fuzzy import FuzzyText, FuzzyInteger

from ..services import Redis

from rorm.client import Client
from rorm.types import String


rlen = FuzzyInteger(10, 50).fuzz()  # used as string length
rs = FuzzyText().fuzz()  # random string
rpi = FuzzyInteger(1, rlen - 1).fuzz()  # random positive integer
rni = FuzzyInteger(-rlen, -1).fuzz()  # random negative integer
rpio = FuzzyInteger(rlen, 2 * rlen).fuzz()  # random positive integer, out of range
rnio = FuzzyInteger(-2 * rlen, -(rlen + 1)).fuzz()  # random negative integer, out of range
rstr = FuzzyText(length=rlen, chars=string.printable).fuzz()  # random string with random length


def setUpModule():
    Client.instance = Redis('test-service-string')
    Client.get_client().flushall()
    global rstring
    rstring = String(rstr)  # redis string type


def tearDownModule():
    Client.get_client().flushall()
    Client.instance.stop()


class GetItemSpecialMethodTestCase(unittest.TestCase):
    """
    Test String.__getitem__ method
    """

    def test_invalid_index_values(self):
        self.assertRaises(ValueError, lambda: rstring[::0])
        self.assertRaises(TypeError, lambda: rstring[None])
        self.assertRaises(TypeError, lambda: rstring[rs])
        self.assertRaises(IndexError, lambda: rstring[rpio])
        self.assertRaises(IndexError, lambda: rstring[rnio])

    def test_single_index_value(self):
        possible_indice_values = (rpi, rni, -1, 1, 0)
        for key in possible_indice_values:
            with self.subTest(key=key):
                self.assertEqual(rstr[key].encode(), rstring[key])

    def test_slice_values_combinations(self):
        """
        Tests all possible combinations of three argument slice
        """

        possible_step_values = (rpi, rpio, rni, rnio, None)

        possible_indice_values = (rpi, rni, rpio, rnio, 0, None)

        possible_start_stop_slices = tuple(  # possible two arg slices [start:stop]
            itertools.product(possible_indice_values, repeat=2))

        possible_start_stop_step_slices = tuple(  # possible three arg slices [start:stop:step]
            itertools.product(possible_start_stop_slices, possible_step_values))

        for indices, step in possible_start_stop_step_slices:
            slice_ = slice(*indices, step)
            with self.subTest(slice_=slice_):
                self.assertEqual(rstr[slice_].encode(), rstring[slice_])


class LenSpecialMethodTestCase(unittest.TestCase):

    def test___len___method_output(self):
        self.assertEqual(len(rstr), len(rstring))
        self.assertEqual(len(rstr), rlen)


class StrSpecialMethodTestCase(unittest.TestCase):

    def test___str___method_output(self):
        self.assertEqual(str(rstr.encode()), rstring.__str__())


class ContainsSpecialMethodTestCase(unittest.TestCase):

    def test_in_operator_using_whole_string(self):
        self.assertIn(rstr.encode(), rstring)

    def test_in_operator_using_string_fragment(self):
        self.assertIn(rstr[rpi:-rni].encode(), rstring)


class ReversedSpecialMethodTestCase(unittest.TestCase):

    def test_order_of_char_after_using_reversed_method(self):
        for rev_rstr, rev_rstring in zip(reversed(rstr), reversed(rstring)):
            with self.subTest(rev_rstr=rev_rstr, rev_rstring=rev_rstring):
                self.assertEqual(rev_rstr.encode(), rev_rstring)


class CountMethodTestCase(unittest.TestCase):

    def test_count_with_every_char_in_string(self):
        for char in rstr:
            with self.subTest(char=char):
                self.assertEqual(
                    rstr.count(char),
                    rstring.count(char.encode()))

    def test_count_with_random_slice(self):
        rsli = rstr[rpi:-rni]
        self.assertEqual(
            rstr.count(rsli),
            rstring.count(rsli.encode()))


class IndexMethodTestCase(unittest.TestCase):

    def test_index_with_every_char_in_string(self):
        for char in rstr:
            with self.subTest(char=char):
                self.assertEqual(
                    rstr.index(char),
                    rstring.index(char.encode()))

    def test_index_with_random_slice(self):
        rsli = rstr[rpi:-rni]
        self.assertEqual(
            rstr.index(rsli),
            rstring.index(rsli.encode()))

    def test_index_with_non_existent_fragment(self):
        nonex = '\0'
        self.assertNotIn(nonex.encode(), rstring)
        self.assertRaises(
            ValueError, lambda: rstring.index(nonex))

test_cases = (
    IndexMethodTestCase,
    CountMethodTestCase,
    ReversedSpecialMethodTestCase,
    ContainsSpecialMethodTestCase,
    StrSpecialMethodTestCase,
    LenSpecialMethodTestCase,
    GetItemSpecialMethodTestCase,
)


def load(loader, test_cases):
    suite = unittest.TestSuite()
    for test_class in test_cases:
        tests = loader.loadTestsFromTestCase(test_class)
        suite.addTests(tests)
    return suite


def run():
    return load(unittest.defaultTestLoader, test_cases)



import redis
import uuid
import collections

from .client import Client


class RedisDataType:
    def __init__(self, *args, key=None, **kwargs):
        self._client = Client.get_client()
        self._key = self._prepare_key(key)

    @property
    def key(self):
        return self._key

    @staticmethod
    def _prepare_key(key):
        return key if key else uuid.uuid4().hex


class Hash(RedisDataType, collections.MutableMapping):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if data:
            try:
                data = dict(data)
            except (TypeError, ValueError) as e:
                raise ValueError(
                    "'data' argument must be iterable sequence of "
                    "2-element tuple, dict, zip or None. ") from e
            else:
                self._client.hmset(self._key, data)

    def __getitem__(self, key):
        value = self._client.hget(self._key, key)
        if value is None:
            raise KeyError(key)
        return value

    def __setitem__(self, key, value):
        return self._client.hset(self._key, key, value)

    def __delitem__(self, key):
        return self._client.hdel(self._key, key)

    def __iter__(self):
        return iter(self._client.hkeys(self._key))

    def __len__(self):
        return self._client.hlen(self._key)

    def __str__(self):
        return str(self.items(as_dict=True))

    def keys(self):
        return self._client.hkeys(self._key)

    def values(self):
        return self._client.hvals(self._key)

    def items(self, as_dict=False):
        items = self._client.hgetall(self._key)
        if as_dict:
            return items
        return tuple(zip(items.keys(), items.values()))


class Sequence(RedisDataType, collections.Sequence):

    def _get(self):
        """
        Returns whole sequence
        """
        raise NotImplementedError

    def _get_ranges(self, index):
        """
        Returns counterparts of ranges for redis '*range' method.
        """
        data_length = len(self)

        if not isinstance(index, slice):
            if not isinstance(index, int):
                raise TypeError('string indices must be integers')
            if -index > data_length or index >= data_length:
                raise IndexError('string index out of range')
            index = slice(index, index+1 or None, None)

        start, stop, step = index.start, index.stop, index.step

        step = 1 if step is None else step
        start = 0 if start is None else start
        if stop is None:
            stop = len(self) - 1
        elif stop == 0 or stop <= -len(self):
            start = -1
            stop = 0
        else:
            stop = stop - 1

        if step < 0:
            raise IndexError(
                'step indice must be positive integer'
            ) from NotImplementedError

        return start, stop, step

    def __contains__(self, value):
        return value in self._get()

    def __getitem__(self, index):
        return NotImplementedError

    def __str__(self):
        return str(self._get())

    def __len__(self):
        return NotImplementedError

    def index(self, value, start=0, stop=None):
        '''S.index(value, [start, [stop]]) -> integer -- return first index of value.
           Raises ValueError if the value is not present.
        '''

        # this is a modified copy of _collections_abc.Sequence.index
        # the orginal lacks for handling slice indices

        if start is not None and start < 0:
            start = max(len(self) + start, 0)
        if stop is not None and stop < 0:
            stop += len(self)

        i = start
        while (stop is None or i < stop) and i < len(self):
            try:
                if self[i:i+len(value)] == value:
                    return i
            except IndexError:
                break
            i += 1
        raise ValueError('substring not found')

    def count(self, value):
        return self._get().count(value)


class MutableSequence(Sequence, collections.MutableSequence):
    pass


class String(Sequence):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        data and self._client.set(self._key, data)

    def __getitem__(self, *args, **kwargs):
        try:
            start, stop, step = self._get_ranges(*args, **kwargs)
        except IndexError:
            # TODO: OPTIMIZE ITEM GET WITH NEGATIVE STEP INDICE
            result = self._get()
            return result.__getitem__(*args, **kwargs)
        return self._client.getrange(self._key, start, stop)[::step]

    def __len__(self):
        return self._client.strlen(self._key)

    def _get(self):
        return self._client.get(self._key)


class List(MutableSequence):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if data:
            if isinstance(data, (list, tuple)):
                raise ValueError("{} argument must be list or tuple".format(data))
            self._client.lpush(self._key, *data)

    def __getitem__(self, *args, **kwargs):
        start, stop, *_ = super().__getitem__(*args, **kwargs)
        value = self._client.lrange(self._key, start, stop)
        return value

    def __setitem__(self, index, value):
        if not isinstance(index, int):
            # FIXME more verbose error mesage
            raise IndexError
        return self._client.lset(self._key, index, value)

    def __len__(self):
        return self._client.llen(self._key)

    def __delitem__(self, index):
        pivot = uuid.uuid4().hex
        self._client.lset(self._key, index, pivot)
        return self._client.lrem(self._key, pivot, 1)  # andymccurdy/redis-py #145, #162 -- reversed order of the parameters

    def insert(self, index, value):
        pivot = uuid.uuid4().hex
        value_after = self[index][0].decode()
        self[index] = pivot
        self._client.linsert(self._key, 'AFTER', pivot, value_after)
        result = self._client.linsert(self._key, 'BEFORE', pivot, value)
        self._client.lrem(self._key, pivot, 1)
        return result
